/*
 * Copyright (C), 2000-2004 by the monit project group.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef HAVE_SOL_IP
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#endif

#ifdef HAVE_NETINET_IP_ICMP_H
#include <netinet/ip_icmp.h>
#endif

#include "monitor.h"
#include "alert.h"
#include "event.h"
#include "socket.h"
#include "net.h"
#include "device.h"
#include "process.h"


/**
 *  Implementation of validation engine
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Olivier Beyssac, <ob@r14.freenix.org> (check_skip)
 *  @author Martin Pala <martinp@tildeslash.com>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: validate.c,v 1.112 2004/04/01 21:08:20 martinp Exp $
 *  @file
 */


/* -------------------------------------------------------------- Prototypes */


static void check_process_state(Service_T);
static void check_process_resources(Service_T, Resource_T);
static void check_process_connection(Service_T, Port_T);
static void check_icmp_connection(Service_T, Icmp_T);
static void check_device_resources(Service_T, Device_T);
static void check_timestamp(Service_T);
static void check_checksum(Service_T);
static void check_perm(Service_T, mode_t);
static void check_uid(Service_T, uid_t);
static void check_gid(Service_T, gid_t);
static void check_size(Service_T, unsigned long);
static int  check_skip(Service_T);
static int  check_timeout(Service_T);
static int  compare_value(int, int, int);


/* ---------------------------------------------------------------- Public */


/**
 *  This function contains the main check machinery for  monit. The
 *  validate function check services in the service list to see if
 *  they will pass all defined tests.
 */
void validate() {

  Service_T s;
  sigset_t ns, os;

  if(! update_loadavg())
    log("Update of loadavg has failed!\n");

  if(Run.doprocess)
    initprocesstree();

  for(s= servicelist; s; s= s->next) {
    if(s->visited)
      continue;
    LOCK(s->mutex)
      set_signal_block(&ns, &os);
      if(s->do_monitor && !check_skip(s) && !check_timeout(s))
        s->check(s);
      unset_signal_block(&os);
    END_LOCK;
  }

  if(Run.doprocess)
    delprocesstree();

  reset_depend();

}


/**
 * Validate a given process service s. Events are posted according to 
 * its configuration. In case of a fatal event FALSE is returned.
 */
int check_process(Service_T s) {

  pid_t  pid= -1;
  Port_T pp= NULL;
  Resource_T pr= NULL;

  ASSERT(s);

  /* Test for running process */
  if(!(pid= is_process_running(s))) {
    /* Reset the proc info object to prevent false data in the first run */
    memset(s->procinfo, 0, sizeof *(s->procinfo));
    Event_post(s, EVENT_NONEXIST, TRUE, s->action_NONEXIST,
      "Process '%s' is not running", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_NONEXIST, FALSE, s->action_NONEXIST,
      "Process '%s' is running with pid %d", s->name, (int)pid);
  }

  if(Run.doprocess) {
    if(update_process_data(s, ptree, ptreesize, pid)) {
      check_process_state(s);
      for(pr= s->resourcelist; pr; pr= pr->next) {
        check_process_resources(s, pr);
      }
    } else {
      log("'%s' failed to get service data\n", s->name);
    }
  }

  /* Test each host:port and protocol in the service's portlist */
  if(s->portlist)
    for(pp= s->portlist; pp; pp= pp->next)
      check_process_connection(s, pp);

  return TRUE;
  
}


/**
 * Validate a given device service s.  Events are posted according to 
 * its configuration.  In case of a fatal event FALSE is returned.
 */
int check_device(Service_T s) {

  Device_T td;
  struct stat stat_buf;

  ASSERT(s);

  if(stat(s->path, &stat_buf) != 0) {
    Event_post(s, EVENT_NONEXIST, TRUE, s->action_NONEXIST,
      "device '%s' doesn't exist", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_NONEXIST, FALSE, s->action_NONEXIST,
      "device '%s' exist", s->name);
  }

  if(s->perm)
    check_perm(s, stat_buf.st_mode);

  if(s->uid)
    check_uid(s, stat_buf.st_uid);

  if(s->gid)
    check_gid(s, stat_buf.st_gid);

  if(!DeviceInfo_Usage(s->devinfo, s->path)) {
    Event_post(s, EVENT_DATA, TRUE, s->action_DATA,
      "Event: unable to read device '%s' state\n", s->path);
    return FALSE;
  } else {
    Event_post(s, EVENT_DATA, FALSE, s->action_DATA,
      "'%s' succeeded getting device statistic for %s", s->name, s->path);
  }

  if(s->devicelist)
    for(td= s->devicelist; td; td= td->next)
      check_device_resources(s, td);

  return TRUE;

}


/**
 * Validate a given file service s.  Events are posted according to 
 * its configuration.  In case of a fatal event FALSE is returned.
 */
int check_file(Service_T s) {

  struct stat stat_buf;

  ASSERT(s);

  if(stat(s->path, &stat_buf) != 0) {
    Event_post(s, EVENT_NONEXIST, TRUE, s->action_NONEXIST,
      "file '%s' doesn't exist", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_NONEXIST, FALSE, s->action_NONEXIST,
      "file '%s' exist", s->name);
  }

  if(!S_ISREG(stat_buf.st_mode)) {
    Event_post(s, EVENT_INVALID, TRUE, s->action_INVALID,
      "'%s' is not regular file", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_INVALID, FALSE, s->action_INVALID,
      "'%s' is regular file", s->name);
  }

  if(s->checksum)
    check_checksum(s);

  if(s->perm)
    check_perm(s, stat_buf.st_mode);

  if(s->uid)
    check_uid(s, stat_buf.st_uid);

  if(s->gid)
    check_gid(s, stat_buf.st_gid);

  if(s->sizelist)
    check_size(s, stat_buf.st_size);

  if(s->timestamplist)
    check_timestamp(s);

  return TRUE;

}


/**
 * Validate a given directory service s. Events are posted according to
 * its configuration. In case of a fatal event FALSE is returned.
 */
int check_directory(Service_T s) {

  struct stat stat_buf;

  ASSERT(s);

  if(stat(s->path, &stat_buf) != 0) {
    Event_post(s, EVENT_NONEXIST, TRUE, s->action_NONEXIST,
      "directory '%s' doesn't exist", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_NONEXIST, FALSE, s->action_NONEXIST,
      "directory '%s' exist", s->name);
  }

  if(!S_ISDIR(stat_buf.st_mode)) {
    Event_post(s, EVENT_INVALID, TRUE, s->action_INVALID,
      "'%s' is not directory", s->name);
    return FALSE;
  } else {
    Event_post(s, EVENT_INVALID, FALSE, s->action_INVALID,
      "'%s' is directory", s->name);
  }

  if(s->perm)
    check_perm(s, stat_buf.st_mode);

  if(s->uid)
    check_uid(s, stat_buf.st_uid);

  if(s->gid)
    check_gid(s, stat_buf.st_gid);

  if(s->timestamplist)
    check_timestamp(s);

  return TRUE;

}


/**
 * Validate a remote service.
 * @param s The remote service to validate
 * @return FALSE if there was an error otherwise TRUE
 */
int check_remote_host(Service_T s) {

  Port_T p= NULL;
  Icmp_T icmp;

  ASSERT(s);

  /* Test each icmp type in the service's icmplist */
  if(s->icmplist) {

    for(icmp= s->icmplist; icmp; icmp= icmp->next)
      check_icmp_connection(s, icmp);

    /* If we could not ping the host we assume it's down and do not
     * continue to check any port connections  */
    for(icmp= s->icmplist; icmp; icmp= icmp->next) {
      if(!icmp->is_available) {
        DEBUG("'%s' icmp ping failed, skipping any port connection tests\n",
          s->name);
        return FALSE;
      }
    }
  }

  /* Test each host:port and protocol in the service's portlist */
  if(s->portlist)
    for(p= s->portlist; p; p= p->next)
      check_process_connection(s, p);

  return TRUE;
  
}


/* --------------------------------------------------------------- Private */


/**
 * Test the connection and protocol
 */
static void check_process_connection(Service_T s, Port_T p) {

  Socket_T socket;
  volatile int rv= TRUE;
  char report[STRLEN]={0};

  ASSERT(s && p);

  /* Open a socket to the destination INET[hostname:port] or UNIX[pathname] */
  socket= socket_create(p);
  if(!socket) {
    snprintf(report, STRLEN,
             "'%s' failed, cannot open a connection to %s",
             s->name, p->address);
    rv= FALSE;
    goto error;
  } else {
    DEBUG("'%s' succeeded connecting to %s\n", s->name, p->address);
  }

  /* Verify that the socket is ready for i|o */
  if(! socket_is_ready(socket)) {
    snprintf(report, STRLEN,
             "'%s' failed, the socket at %s is not ready for i|o -- %s",
             s->name, p->address, STRERROR);
    rv= FALSE;
    goto error;
  }

  /* Run the protocol verification routine through the socket */
  if(! p->protocol->check(socket)) {
    snprintf(report, STRLEN,
             "'%s' failed protocol test [%s] at %s.",
             s->name, p->protocol->name, p->address);
    rv= FALSE;
    goto error;
  } else {
    DEBUG("'%s' succeeded testing protocol [%s] at %s\n",
          s->name, p->protocol->name, p->address);
  }

  error:
  if(socket) socket_free(&socket);

  if(!rv) {
    p->is_available= FALSE;
    Event_post(s, EVENT_CONNECTION, TRUE, p->action, report);
  } else {
    p->is_available= TRUE;
    Event_post(s, EVENT_CONNECTION, FALSE, p->action,
      "'%s' connection passed", s->name);
  }
      
}


/**
 * Test ICMP connection
 */
static void check_icmp_connection(Service_T s, Icmp_T i) {

  ASSERT(s && i);

  if(i->type == ICMP_ECHO && !icmp_echo(s->path, i->timeout)) {
    i->is_available= FALSE;
    Event_post(s, EVENT_CONNECTION, TRUE, i->action,
      "'%s' failed ICMP test [%s]", s->name, icmpnames[i->type]);
  } else {
    DEBUG("'%s' succeeded testing ICMP [%s]\n", s->name, icmpnames[i->type]);
    i->is_available= TRUE;
    Event_post(s, EVENT_CONNECTION, FALSE, i->action,
      "'%s' icmp passed", s->name);
  }

}


/**
 * Test process state (e.g. Zombie)
 */
static void check_process_state(Service_T s) {

  ProcInfo_T pi;

  ASSERT(s);

  pi= s->procinfo;

  if(pi->status_flag & PROCESS_ZOMBIE) {
    /* We do not check the process anymore if it's a zombie
       since such a process is (usually) unmanageable */
    LOCK(Run.mutex)
      s->do_monitor= FALSE;
    END_LOCK;
    Event_post(s, EVENT_DATA, TRUE, s->action_DATA,
      "process with pid %d is a zombie", pi->pid);
  } else {
    DEBUG("'%s' zombie check passed [status_flag=%04x]\n",
      s->name,  pi->status_flag);
    Event_post(s, EVENT_DATA, FALSE, s->action_DATA,
      "'%s' check process state passed", s->name);
  }

}


/**
 * Returns TRUE if the proc test succeeded
 * otherwise FALSE.
 */
static void check_process_resources(Service_T s, Resource_T pr) {

  ProcInfo_T pi;
  int okay= TRUE;
  char report[STRLEN]={0};

  ASSERT(s && pr);

  pi= s->procinfo;

  switch(pr->resource_id) {
  case RESOURCE_ID_CPU_PERCENT:
      if(compare_value(pr->operator, pi->cpu_percent, pr->limit)) {
        snprintf(report, STRLEN,
          "cpu usage of %.1f%% matches resource limit [cpu usage%s%.1f%%]",
          pi->cpu_percent/10.0, operatorshortnames[pr->operator],
          pr->limit/10.0);
        okay= FALSE;
      } else {
        DEBUG("'%s' cpu usage check passed [current cpu usage=%.1f%%]\n",
          s->name, pi->cpu_percent/10.0);
      }
      break;

  case RESOURCE_ID_MEM_PERCENT:
      if(compare_value(pr->operator, pi->mem_percent, pr->limit)) {
        snprintf(report, STRLEN,
          "mem usage of %.1f%% matches resource limit [mem usage%s%.1f%%]",
          pi->mem_percent/10.0, operatorshortnames[pr->operator],
          pr->limit/10.0);
        okay= FALSE;
      } else {
        DEBUG("'%s' mem usage check passed [current mem usage=%.1f%%]\n",
          s->name, pi->mem_percent/10.0);
      }
      break;

  case RESOURCE_ID_MEM_KBYTE:
    if(compare_value(pr->operator, pi->mem_kbyte, pr->limit)) {
      snprintf(report, STRLEN,
        "mem amount of %ldkB matches resource limit [mem amount%s%ldkB]",
        pi->mem_kbyte, operatorshortnames[pr->operator], pr->limit);
      okay= FALSE;
    } else {
      DEBUG("'%s' mem amount check passed [current mem amount=%ldkB]\n",
        s->name, pi->mem_kbyte);
    }
    break;

  case RESOURCE_ID_LOAD1:
    if(compare_value(pr->operator, (int)(Run.loadavg[0]*10.0), pr->limit)) {
      snprintf(report, STRLEN,
        "loadavg(1min) of %.1f matches resource limit [loadavg(1min)%s%.1f]",
        Run.loadavg[0], operatorshortnames[pr->operator], pr->limit/10.0);
      okay= FALSE;
    } else {
      DEBUG("'%s' loadavg(1min) check passed [current loadavg(1min)=%.1f]\n",
        s->name, Run.loadavg[0]);
    }
    break;

  case RESOURCE_ID_LOAD5:
    if(compare_value(pr->operator, (int)(Run.loadavg[1]*10.0), pr->limit)) {
      snprintf(report, STRLEN,
        "loadavg(5min) of %.1f matches resource limit [loadavg(5min)%s%.1f]",
        Run.loadavg[1], operatorshortnames[pr->operator], pr->limit/10.0);
      okay= FALSE;
    } else {
      DEBUG("'%s' loadavg(5min) check passed [current loadavg(5min)=%.1f]\n",
        s->name, Run.loadavg[1]);
    }
    break;

  case RESOURCE_ID_LOAD15:
      if(compare_value(pr->operator, (int)(Run.loadavg[2]*10.0), pr->limit)) {
        snprintf(report, STRLEN,
          "loadavg(15min) of %.1f matches resource limit "
          "[loadavg(15min)%s%.1f]",
          Run.loadavg[2], operatorshortnames[pr->operator], pr->limit/10.0);
        okay= FALSE;
      } else {
        DEBUG("'%s' loadavg(15min) check passed [current loadavg(15min)=%.1f]\n",
          s->name, Run.loadavg[2]);
      }
      break;

  case RESOURCE_ID_CHILDREN:
      if(compare_value(pr->operator, pi->children, pr->limit)) {
        snprintf(report, STRLEN,
          "children of %i matches resource limit [children%s%ld]",
          pi->children, operatorshortnames[pr->operator], pr->limit);
        okay= FALSE;
      } else {
        DEBUG("'%s' children check passed [current children=%i]\n",
          s->name, pi->children);
      }
    break;

  case RESOURCE_ID_TOTAL_MEM_KBYTE:
    if(compare_value(pr->operator, pi->total_mem_kbyte, pr->limit)) {
      snprintf(report, STRLEN,
        "total mem amount of %ldkB matches resource limit"
        " [total mem amount%s%ldkB]",
        pi->total_mem_kbyte, operatorshortnames[pr->operator], pr->limit);
      okay= FALSE;
    } else {
      DEBUG("'%s' total mem amount check passed "
        "[current total mem amount=%ldkB]\n", s->name, pi->total_mem_kbyte);
    }
    break;

  default:
    log("'%s' error -- unknown resource ID: [%d]\n", s->name, pr->resource_id);
  }

  if(okay && pr->cycle > 0) {
    pr->cycle--;
  } else if(! okay) {
    pr->cycle++;
  }

  if(pr->cycle >= pr->max_cycle) {
    pr->cycle=0;
    Event_post(s, EVENT_RESOURCE, TRUE, pr->action, report);
  } else {
    Event_post(s, EVENT_RESOURCE, FALSE, pr->action,
      "'%s' resource passed", s->name);
  }
  
}


/**
 * Returns TRUE if the service timed out, otherwise FALSE.
 */
static int check_timeout(Service_T s) {

  ASSERT(s);

  if(!s->def_timeout)
    return FALSE;

  /*
   * Start counting cycles
   */
  if(s->nstart > 0)
    s->ncycle++;

  /*
   * Check timeout
   */
  if(s->nstart >= s->to_start && s->ncycle <= s->to_cycle) {
    Event_post(s, EVENT_TIMEOUT, TRUE, s->action_TIMEOUT,
              "service '%s' timed out and will not be checked anymore",
              s->name);
    return TRUE;
  }

  /*
   * Stop counting and reset if the
   * cycle interval is passed
   */
  if(s->ncycle > s->to_cycle) {
    s->ncycle= 0;
    s->nstart= 0;
  }

  return FALSE;

}


/**
 * Returns TRUE if validation should be skiped for
 * this service in this cycle, otherwise FALSE
 */
static int check_skip(Service_T s) {

  ASSERT(s);
  
  if(!s->def_every)
    return FALSE;
  
  if(++s->nevery < s->every)
    return TRUE;

  s->nevery= 0;

  return FALSE;

}


/**
 * Test for associated path checksum change
 */
static void check_checksum(Service_T s) {

  Checksum_T cs;

  ASSERT(s && s->checksum);

  cs= s->checksum;

  if( !check_hash(s->path, cs->hash, cs->type) ) {

    /* if we are testing for changes only, the value is variable */
    if(cs->test_changes) {
        Event_post(s, EVENT_CHANGED, TRUE, cs->action,
          "checksum was changed for %s", s->path);
        /* reset value */
        FREE(cs->hash);
        cs->hash= get_checksum(s->path, cs->type);
      /* the EVENT_CHANGED has no inversion possible => return */
      return;
    }

    /* we are testing constant value for failed or passed state */
    Event_post(s, EVENT_CHECKSUM, TRUE, cs->action,
      "checksum test failed for %s", s->path);

  } else if(cs->test_changes) {
      DEBUG("'%s' checksum has not changed\n", s->name);
  } else {
      DEBUG("'%s' has valid checksums\n", s->name);
      Event_post(s, EVENT_CHECKSUM, FALSE, cs->action,
        "'%s' checksum passed", s->name);
  }

}


/**
 * Test for associated path permission change
 */
static void check_perm(Service_T s, mode_t mode) {

  ASSERT(s && s->perm);

  if((mode & 07777) != s->perm->perm) {
    Event_post(s, EVENT_PERMISSION, TRUE, s->perm->action,
      "permission test failed for %s -- current permission is %o",
      s->path, mode&07777);
  } else {
    DEBUG("'%s' file permission check passed [current permission=%o]\n",
      s->name, mode&07777);
    Event_post(s, EVENT_PERMISSION, FALSE, s->perm->action,
      "'%s' permission passed", s->name);
  }
}


/**
 * Test for associated path uid change
 */
static void check_uid(Service_T s, uid_t uid) {

  ASSERT(s && s->uid);

  if(uid != s->uid->uid) {
    Event_post(s, EVENT_UID, TRUE, s->uid->action,
      "uid test failed for %s -- current uid is %d", s->path, (int)uid);
  } else {
    DEBUG("'%s' device uid check passed [current uid=%d]\n", s->name, (int)uid);
    Event_post(s, EVENT_UID, FALSE, s->uid->action, "'%s' uid passed", s->name);
  }

}


/**
 * Test for associated path gid change
 */
static void check_gid(Service_T s, gid_t gid) {

  ASSERT(s && s->gid);

  if(gid != s->gid->gid ) {
    Event_post(s, EVENT_GID, TRUE, s->gid->action,
      "gid test failed for %s -- current gid is %d", s->path, (int)gid);
  } else {
    DEBUG("'%s' device gid check passed [current gid=%d]\n", s->name, (int)gid);
    Event_post(s, EVENT_GID, FALSE, s->gid->action, "'%s' gid passed", s->name);
  }

}


/**
 * Validate timestamps of a service s
 */
static void check_timestamp(Service_T s) {

  time_t now;
  time_t timestamp;
  Timestamp_T t;

  ASSERT(s && s->timestamplist);

  for(t= s->timestamplist; t; t= t->next) {
    if((int)time(&now) == -1) {
      Event_post(s, EVENT_TIMESTAMP, TRUE, t->action, "can't get actual time");
      break;
    }

    if(!(timestamp= get_timestamp(s->path, S_IFDIR|S_IFREG))) {
      Event_post(s, EVENT_TIMESTAMP, TRUE, t->action,
        "can't get timestamp for %s", s->path);
      break;
    }

    /* if we are testing for changes only, the value is variable */
    if(t->test_changes) {
      if(t->timestamp != timestamp) {
        Event_post(s, EVENT_CHANGED, TRUE, t->action,
          "timestamp was changed for %s", s->path);
        /* reset value */
        t->timestamp= timestamp;
      } else {
        DEBUG("'%s' timestamp has not changed for %s\n", s->name, s->path);
      }
      /* the EVENT_CHANGED has no inversion possible => return from the cycle */
      break;
    }

    /* we are testing constant value for failed or passed state */
    if(compare_value(t->operator, (int)(now - timestamp), t->time)) {
      Event_post(s, EVENT_TIMESTAMP, TRUE, t->action,
        "timestamp test failed for %s", s->path);
    } else {
      DEBUG("'%s' timestamp test passed for %s\n", s->name, s->path); 
      Event_post(s, EVENT_TIMESTAMP, FALSE, t->action,
        "'%s' timestamp passed", s->name);
    }
  }
}


/**
 * Test size
 */
static void check_size(Service_T s, unsigned long size) {

  Size_T sl;

  ASSERT(s && s->sizelist);

  for(sl= s->sizelist; sl; sl= sl->next) {

    /* if we are testing for changes only, the value is variable */
    if(sl->test_changes) {
      if(sl->size != size) {
        Event_post(s, EVENT_CHANGED, TRUE, sl->action,
          "size was changed for %s", s->path);
        /* reset value */
        sl->size= size;
      } else {
        DEBUG("'%s' size has not changed [current size=%lu B]\n", s->name, size);
      }
      /* the EVENT_CHANGED has no inversion possible => return from the cycle */
      break;
    }

    /* we are testing constant value for failed or passed state */
    if(compare_value(sl->operator, size, sl->size)) {
      Event_post(s, EVENT_SIZE, TRUE, sl->action,
        "size test failed for %s -- current size is %lu B", s->path, size);
    } else {
      DEBUG("'%s' file size check passed [current size=%lu B]\n", s->name, size);
      Event_post(s, EVENT_SIZE, FALSE, sl->action, "'%s' size passed", s->name);
    }
  }
}


/**
 * Device test
 */
static void check_device_resources(Service_T s, Device_T td) {

  ASSERT(s && td);

  if( (td->limit_percent < 0) && (td->limit_absolute < 0) ) {
    log("'%s' error: device limit not set\n", s->name);
    return;
  }

  switch(td->resource) {

  case RESOURCE_ID_INODE:
      if(s->devinfo->f_files <= 0) {
	DEBUG("'%s' filesystem doesn't support inodes\n", s->name);
	return;
      }

      if(td->limit_percent >= 0) {
	if(compare_value(
	       td->operator,
	       100 * (s->devinfo->f_files - s->devinfo->f_filesfree) /
	       s->devinfo->f_files, td->limit_percent)) {
          Event_post(s, EVENT_RESOURCE, TRUE, td->action,
            "inode usage %ld%% matches resource limit [inode usage%s%d%%]",
	    100 * (s->devinfo->f_files - s->devinfo->f_filesfree) /
	    s->devinfo->f_files, operatorshortnames[td->operator],
	    td->limit_percent);
	  return;
	}
      } else {
	if(compare_value(td->operator, s->devinfo->f_files -
			 s->devinfo->f_filesfree, td->limit_absolute)) {
          Event_post(s, EVENT_RESOURCE, TRUE, td->action,
            "inode usage %ld matches resource limit [inode usage%s%ld]",
            s->devinfo->f_files - s->devinfo->f_filesfree,
            operatorshortnames[td->operator], td->limit_absolute);
	  return;
	}
      }
      DEBUG("'%s' inode usage check passed [current inode usage=%.1f%%]\n",
	    s->name,
	    (float) 100 * (s->devinfo->f_files - s->devinfo->f_filesfree) /
	    s->devinfo->f_files);
      Event_post(s, EVENT_RESOURCE, FALSE, td->action,
        "'%s' device resources passed", s->name);
      return;

  case RESOURCE_ID_SPACE:
      if(td->limit_percent >= 0) {
        if(compare_value(
	       td->operator,
	       100 * (s->devinfo->f_blocks - s->devinfo->f_blocksfreetotal) /
	       s->devinfo->f_blocks, td->limit_percent)) {
          Event_post(s, EVENT_RESOURCE, TRUE, td->action,
	    "space usage %ld%% matches resource limit [space usage%s%d%%]",
            100*(s->devinfo->f_blocks - s->devinfo->f_blocksfreetotal) /
            s->devinfo->f_blocks, operatorshortnames[td->operator],
            td->limit_percent);
          return;
        }
      } else {
        if(compare_value(td->operator, s->devinfo->f_blocks -
			 s->devinfo->f_blocksfreetotal, td->limit_absolute)) {
          Event_post(s, EVENT_RESOURCE, TRUE, td->action,
            "space usage %ld blocks matches resource limit "
            "[space usage%s%ld blocks]",
            s->devinfo->f_blocks - s->devinfo->f_blocksfreetotal,
            operatorshortnames[td->operator], td->limit_absolute);
	  return;
        }
      }
      DEBUG("'%s' space usage check passed [current space usage=%.1f%%]\n",
	    s->name,
	    (float) 100*(s->devinfo->f_blocks - s->devinfo->f_blocksfreetotal) /
	    s->devinfo->f_blocks);
      Event_post(s, EVENT_RESOURCE, FALSE, td->action,
        "'%s' device resources passed", s->name);
      return;
      
  default:
      log("'%s' error -- unknown resource type: [%d]\n", s->name, td->resource);
      return;
  }
  
}


/**
 * Comparison of values. Returns TRUE if comparison matches, otherwise
 * FALSE.
 */
static int compare_value(int operator, int left, int right) {

  switch(operator) {
  case OPERATOR_GREATER:
      if(left > right)
	  return TRUE;
      break;
  case OPERATOR_LESS:
      if(left < right)
	  return TRUE;
      break;
  case OPERATOR_EQUAL:
      if(left == right)
	  return TRUE;
      break;
  case OPERATOR_NOTEQUAL:
      if(left != right)
	  return TRUE;
      break;
  default:
      log("Unknown comparison operator\n");
      return FALSE;
  }

  return FALSE;
    
}

